// Copyright (c) 2005..2010 Dirk Jagdmann <doj@cubic.org>
//
// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.
//
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
//
//     1. The origin of this software must not be misrepresented; you
//        must not claim that you wrote the original software. If you use
//        this software in a product, an acknowledgment in the product
//        documentation would be appreciated but is not required.
//
//     2. Altered source versions must be plainly marked as such, and
//        must not be misrepresented as being the original software.
//
//     3. This notice may not be removed or altered from any source
//        distribution.

// $Header: /code/doj/file.hpp,v 1.18 2010/01/04 02:40:18 doj Exp $

#ifndef FILE__HPP
#define FILE__HPP

// RAII wrapper for stdio FILE pointer
// from Bjarne Stroustrup "The C++ Programming Language"
// then enhanced by Dirk Jagdmann <doj@cubic.org>

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <unistd.h>
#include <cstdio>
#include <string>

namespace doj
{

  /**
     \todo set a flag to use function from unlocked_stdio(3)
   */
  class FILEp
  {
    FILE *fp_;			///< pointer to internal FILE object
  public:
    /** constructor from existing FILE object. The constructed object
	of FILEp takes ownership of fp.
	@param fp pointer to FILE object
    */
    explicit FILEp(FILE *fp) : fp_(fp) { }

    /** constructor which opens/creates a file. For details on mode see fopen(3).
	@param filename pathname of file in local filesystem
	@param mode mode in which filename is accessed, allowed values are "r", "r+", "w", "w+", "a", "a+"
     */
    explicit FILEp(const char * const filename, const char * const mode) : fp_(fopen(filename, mode)) { }

    /** constructor which opens/creates a file. For details on mode see fopen(3).
	@param filename pathname of file in local filesystem
	@param mode mode in which filename is accessed, allowed values are "r", "r+", "w", "w+", "a", "a+"
     */
    explicit FILEp(const std::string& filename, const std::string& mode) : fp_(fopen(filename.c_str(), mode.c_str())) { }

    /** constructor which opens/creates a file. For details on mode see fopen(3).
	@param filename pathname of file in local filesystem
	@param mode mode in which filename is accessed, allowed values are "r", "r+", "w", "w+", "a", "a+"
     */
    explicit FILEp(const std::string& filename, const char * const mode) : fp_(fopen(filename.c_str(), mode)) { }

    ~FILEp() { if(fp_) fclose(fp_); }

    //lint -e{1763,1930}
    /// @return pointer to (internal) FILE object
    operator FILE*() const { return fp_; }

    //lint -e{1763}
    /// @return pointer to (internal) FILE object
    FILE* get() const { return fp_; }

    /** releases the ownership of the internal FILE object from
	this. The caller then has to fclose the returned object
	(pointer) or construct another FILEp object with it.
	@return pointer to (internal) FILE object
    */
    FILE* release()
    {
      //lint --e{953} ignore non const temp
      FILE *temp=fp_;
      fp_=0;
      return temp;
    }

    /** swaps this with a.
	@param[in,out] a another FILEp object
	@return this
    */
    FILEp& swap(FILEp& a)
    {
      //lint --e{953} ignore non const t
      FILE *t=fp_;
      fp_=a.fp_;
      a.fp_=t;
      return *this;
    }

    //lint --e{713} ignore loss of precision

    /** flushes any (write) buffers. For details see fflush(3)
	@return 0 upon success, EOF upon error.
    */
    int flush() { if(!fp_) return -1; return fflush(fp_); }

    /// clear error status. For details see clearerr(3)
    void clearerr() { if(fp_) ::clearerr(fp_); }

    /** tests if the end of file has been reached. For details see feof(3)
	@return non-zero if end of file reached
    */
    int eof() { if(!fp_) return -1; return feof(fp_); }

    /** tests for stream error status. For details see ferror(3)
	@return non-zero if the stream has an error state
    */
    int error() { if(!fp_) return -1; return ferror(fp_); }

    /// @return file descriptor of stream upon success, negative upon error
    int fd() { if(!fp_) return -1; return fileno(fp_); }

    /** read from stream. For details see fread(3)
	@param ptr pointer to buffer of size*nmemb bytes size
	@param size size of one "member" in ptr in bytes, must be > 0
	@param nmemb number of "members" in ptr, must be > 0
	@return number of members read, <=0 upon error
    */
    int read(void *ptr, size_t size, size_t nmemb) { if(!fp_) return -1; return fread(ptr, size, nmemb, fp_); }

    /** write stream. For details see fread(3)
	@param ptr pointer to buffer of size*nmemb bytes size
	@param size size of one "member" in ptr in bytes, must be > 0
	@param nmemb number of "members" in ptr, must be > 0
	@return number of members written, <=0 upon error
    */
    int write(const void * const ptr, size_t size, size_t nmemb) { if(!fp_) return -1; return fwrite(ptr, size, nmemb, fp_); }

    /** get file status. For details see fstat(3)
	@param buf pointer to struct stat
	@return 0 upon success, negative upon error
    */
    int stat(struct stat *buf) { if(!fp_) return -1; return fstat(fd(), buf); }

    /// @return position in stream in bytes, negative upon error. For details see ftell(3)
    off_t tell() { if(!fp_) return -1; return ftello(fp_); }

    /// set stream position to beginning (0). For details see rewind(3)
    void rewind() { if(fp_) ::rewind(fp_); }

    /** position in stream. For details see fseek(3)
	@param offset offset from ancor (whence)
	@param whence SEEK_SET, SEEK_CUR, or SEEK_END
	@return 0 upon success, negative upon error
    */
    int seek(off_t offset, int whence) { if(!fp_) return -1; return fseeko(fp_, offset, whence); }

    /** write a string to the stream.
	@param s string to write
	@return number of bytes written, <=0 upon error
    */
    int write(const std::string& s) { return write(s.data(), 1, s.size()); }

    /** read a character from the stream. For details see fgetc(3)
	@return character value (0-255), negative upon error
    */
    int get() { if(!fp_) return EOF; return fgetc(fp_); }

    /** read a line from the stream. For details see fgets(3)
	@param s pointer to buffer of size bytes
	@param size length of s in bytes
	@return s upon success, NULL upon error
    */
    char* gets(char *s, int size) { if(!fp_) return NULL; return fgets(s, size, fp_); }

    /** close the stream. For details see fclose(3)
	@return 0 upon success, negative upon error
    */
    int close() { if(!fp_) return -1; const int ret=fclose(fp_); fp_=0; return ret; }

    /** write a character to the stream. For details see fputc(3)
	@param c character value (0-255)
	@return c upon success, negative upon error
    */
    int put(int c) { if(!fp_) return -1; return fputc(c, fp_); }

    /** write s with trailing newline. This is more like puts() than fputs()!
	@param s string
	@return negative upon error.
    */
    int puts(std::string s)
    {
      if(!fp_)
	return -1;
      if(s.empty()) s="\n";
      else if(s[s.size()-1]!='\n') s+='\n';
      return fputs(s.c_str(), fp_);
    }

  private:
    //lint --e{1704} we want those constructors private
    FILEp();
    FILEp(const FILEp&);
    FILEp& operator=(const FILEp&);
  };

}

#endif

#ifdef TESTFILE
#include <errno.h>
#define TMPFILE "/tmp/file.hpp.test"
#define TESTSTR "Dies ist der Test"
int main()
{
  {
    doj::FILEp f(TMPFILE, "w");
    if(f)
      {
	f.puts(TESTSTR);
	fprintf(f, "%i\n", getpid());
      }
    else
      {
	fprintf(stderr, "could not open %s: %s\n", TMPFILE, strerror(errno));
	return 1;
      }
  }
  {
    doj::FILEp f(TMPFILE, "r");
    if(f)
      {
	char buf[100];
	f.gets(buf, sizeof(buf));
	if(strcmp(buf, TESTSTR "\n"))
	  {
	    fprintf(stderr, "file test failed!\n");
	    return 1;
	  }
      }
    else
      {
	fprintf(stderr, "could not open %s: %s\n", TMPFILE, strerror(errno));
	return 1;
      }
  }

  printf("file test success.\n");
  return 0;
}
#endif
