// Copyright (c) 2010 Dirk Jagdmann <doj@cubic.org>

// This software is provided 'as-is', without any express or implied
// warranty. In no event will the authors be held liable for any damages
// arising from the use of this software.

// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:

//     1. The origin of this software must not be misrepresented; you
//        must not claim that you wrote the original software. If you use
//        this software in a product, an acknowledgment in the product
//        documentation would be appreciated but is not required.

//     2. Altered source versions must be plainly marked as such, and
//        must not be misrepresented as being the original software.

//     3. This notice may not be removed or altered from any source
//        distribution.

#include <errno.h>
#include <getopt.h>
#include <sysexits.h>
#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <limits>
using namespace std;
#include "memorymap.hpp"
#include "memorymap.cpp"
#include "wav.h"

void help()
{
  cout << "usage: wavmerge -o <outfile> [infile]*" << endl;
  cout << "written by Dirk Jagdmann <doj@cubic.org> http://llg.cubic.org" << endl;
  exit(EX_USAGE);
}

FILE *out=NULL;
char *outfile=0;

void exit_unlink(int e)
{
  if(out)
    fclose(out);
  if(outfile)
    unlink(outfile);
  exit(e);
}

int main(int argc, char **argv)
{
  int c=1;

  while((c=getopt(argc, argv, "o:h")) != -1)
    switch(c)
      {
      case 'o': outfile=optarg; break;
      case 'h':
      default:  help();
      }

  if(outfile==0)
    help();

  out=fopen(outfile, "wb");
  if(!out)
    {
      cerr << "could not open " << outfile << ":" << strerror(errno) << endl;
      exit(EX_NOINPUT);
    }

  WAV w;

  uint64_t len=0;
  for(int i=optind; i<argc; i++)
    {
      WAV *wav=(WAV*)doj::memorymap(argv[i]);
      if(wav)
	{
	  if(strncmp(wav->name, "RIFF", 4))
	    {
	      cerr << argv[i] << " is not a RIFF container" << endl;
	      exit_unlink(EX_DATAERR);
	    }
	  if(strncmp(wav->type, "WAVE", 4))
	    {
	      cerr << argv[i] << " is not a WAVE container" << endl;
	      exit_unlink(EX_DATAERR);
	    }
	  if(strncmp(wav->fmt.name, "fmt ", 4))
	    {
	      cerr << argv[i] << " is not a WAVE does not have 'fmt ' chunk" << endl;
	      exit_unlink(EX_DATAERR);
	    }

	  // save first header
	  if(i==optind)
	    {
	      memcpy(&w, wav, sizeof(w));
	      fwrite(wav, sizeof(WAV), 1, out);
	    }
	  // check format
	  else
	    {
	      if(wav->fmt.type != w.fmt.type)
		{
		  cerr << argv[i] << " type does not match first file" << endl;
		  cerr << argv[i] << " type: " << wav->fmt.type << endl;
		  cerr << argv[optind] << " type: " << w.fmt.type << endl;
		  exit_unlink(EX_DATAERR);
		}
	      if(wav->fmt.channels != w.fmt.channels)
		{
		  cerr << argv[i] << " channels do not match first file" << endl;
		  cerr << argv[i] << " channels: " << wav->fmt.channels << endl;
		  cerr << argv[optind] << " channels: " << w.fmt.channels << endl;
		  exit_unlink(EX_DATAERR);
		}
	      if(wav->fmt.rate != w.fmt.rate)
		{
		  cerr << argv[i] << " rate does not match first file" << endl;
		  cerr << argv[i] << " rate: " << wav->fmt.rate << endl;
		  cerr << argv[optind] << " rate: " << w.fmt.rate << endl;
		  exit_unlink(EX_DATAERR);
		}
	      if(wav->fmt.bytesPerSecond != w.fmt.bytesPerSecond)
		{
		  cerr << argv[i] << " bytesPerSecond do not match first file" << endl;
		  cerr << argv[i] << " bytesPerSecond: " << wav->fmt.bytesPerSecond << endl;
		  cerr << argv[optind] << " bytesPerSecond: " << w.fmt.bytesPerSecond << endl;
		  exit_unlink(EX_DATAERR);
		}
	      if(wav->fmt.bitsPerSample != w.fmt.bitsPerSample)
		{
		  cerr << argv[i] << " bytesPerSecond do not match first file" << endl;
		  cerr << argv[i] << " bitsPerSample: " << wav->fmt.bitsPerSample << endl;
		  cerr << argv[optind] << " bitsPerSample: " << w.fmt.bitsPerSample << endl;
		  exit_unlink(EX_DATAERR);
		}
	    }
	  len+=wav->data.length;
	  if(len > std::numeric_limits<uint32_t>::max()-42u)
	    {
	      cerr << "appending " << argv[i] << " would exceed the 4GiB limit of WAVE files" << endl;
	      exit_unlink(EX_DATAERR);
	    }
	  len+=fwrite(wav->data.data(), 1, wav->data.length, out);
	  doj::memoryunmap(wav);
	}
      else
	{
	  cerr << "could not open " << argv[i] << strerror(errno) << endl;
	  exit_unlink(EX_NOINPUT);
	}
    }

  // write correct length to header
  w.setlength(len);

  rewind(out);
  fwrite(&w, sizeof(w), 1, out);
  fclose(out);

  return EX_OK;
}
